@@ -1,11 +1,3 @@
-package Exception::DB;
-$Exception::DB::VERSION = '0.013';
-use base qw(Exception);
-
-package Exception::DB::DuplicateEntry;
-$Exception::DB::DuplicateEntry::VERSION = '0.013';
-use base qw(Exception::DB);
-
 =head1 Name
  
 QBit::Application::Model::DB - base class for DB.
@@ -35,11 +27,14 @@ apt-get install libqbit-application-model-db-perl (http://perlhub.ru/)
 =cut
 
 package QBit::Application::Model::DB;
-$QBit::Application::Model::DB::VERSION = '0.013';
+$QBit::Application::Model::DB::VERSION = '0.014';
 use qbit;
 
 use base qw(QBit::Application::Model);
 
+use Exception::DB;
+use Exception::DB::DuplicateEntry;
+
 use DBI;
 
 =head1 Debug
@@ -431,7 +426,13 @@ sub transaction {
     }
     catch {
         my $ex = shift;
-        $self->rollback();
+
+        unless ($ex->isa('Exception::DB')) {
+            $self->rollback();
+        } else {
+            $self->{'__SAVEPOINTS__'} = 0;
+        }
+
         throw $ex;
     };
 
@@ -536,7 +537,8 @@ sub finish {
     my ($self) = @_;
 
     if ($self->{'__SAVEPOINTS__'}) {
-        $self->rollback() while $self->{'__SAVEPOINTS__'};
+        $self->{'__SAVEPOINTS__'} = 1;
+        $self->rollback();
         throw gettext("Unclosed transaction");
     }
 }
